#include <cmath>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include "randomc.h"
#include "mersenne.cpp"

using namespace std;

typedef vector<double> d_vec;

const double NEVER = 1000000000.0;
const unsigned int NO_JOB = 555;

class job
{ public: 
	unsigned int t;            // type of job        
	vector<double> data;

	job()
	{// data: arrival time, processing time, due date, setup, dispatching rule index   
		data.resize(5);
	}
};

class machine
{public:
	unsigned int machine_state;              // 1: working, 2: idle
	job current_job;                         // current job of working machine 
	unsigned int previous_job_type;          // type of previous job
	unsigned int machine_state_transition;   // 12: if transition from working to idle, 21: if transition from idle to working

	machine()
	{
		machine_state = 2;
		previous_job_type = 0;
		machine_state_transition=0;
	}
};

class parallel_machines 
{ 
public:

unsigned int num_of_machines;              //number of parallel machines
unsigned int num_of_job_types;             //number of job types
double* means_of_times_between_arrivals;   //mean interarrival times of job types
double* means_of_processing_times;         //mean processing times of job types
double* tightness_factors;                 //due date parameters of job types
double* setup_fractions;                   //setup parameters of job types

double t_now;                              //current time
vector<double> times_of_next_events;       
unsigned int event_id;                     //type of event

vector<job> queueing_jobs;                 //jobs pending for processing
vector<machine> machines;                  //parallel machines

unsigned int total_production;               //number of completed jobs
double mean_cycle_time;                    //average cycle time
double mean_earliness;                     //average earliness
double mean_tardiness;                     //average tardiness

parallel_machines(unsigned int some_num_of_job_types, unsigned int some_num_of_machines)
{
	num_of_machines=some_num_of_machines;
	num_of_job_types = some_num_of_job_types;
	means_of_times_between_arrivals = new double[num_of_job_types];
	means_of_processing_times = new double[num_of_job_types];
	tightness_factors = new double[num_of_job_types];
	setup_fractions = new double[num_of_job_types];

	//events are arrival of type i job and production in machine j 
	times_of_next_events.resize(num_of_job_types+num_of_machines);
	event_id=0;
	t_now = 0.0;

	machines.resize(num_of_machines);

	total_production = 0;
	mean_cycle_time=0.0;
	mean_earliness = 0.0;
	mean_tardiness = 0.0;
}

~parallel_machines() 
{
	delete [] means_of_processing_times ;
	delete [] tightness_factors ;
	delete [] setup_fractions ;
	delete [] means_of_times_between_arrivals ;
}

void initialize_instance_from_file()
{// initialize simulation parameters from text file "instance.txt"

	ifstream input_file;     input_file.open("instance.txt",ios::in);

	for(unsigned int i=0;i<num_of_job_types;i++)
		input_file>>means_of_times_between_arrivals[i];     //initialize mean interarrival times of job types

	for(unsigned int i=0;i<num_of_job_types;i++)
		input_file>>means_of_processing_times[i];           //initialize mean processing times of job types

	for(unsigned int i=0;i<num_of_job_types;i++)
		input_file>>tightness_factors[i];                   //initialize due date parameters of job types 

	for(unsigned int i=0;i<num_of_job_types;i++)
		input_file>>setup_fractions[i];                     //initialize setup parameters of job types

	input_file.close();
}

void initialize_times_of_next_events(TRandomMersenne* rand_gen)
{
	for (unsigned int i=0; i<num_of_job_types; i++)
	{//initialize times of next arrivals for each job type
		times_of_next_events[i]=(-1.0)*log(rand_gen->Random())*means_of_times_between_arrivals[i];
	}

	for (unsigned int i=num_of_job_types;i<times_of_next_events.size(); i++)
	{//initialize times of next production for each machine
		times_of_next_events[i]=NEVER;  
	}
}

void find_time_of_next_event()
{ 
double min=(*min_element(times_of_next_events.begin(),times_of_next_events.end()) );
t_now = min;
for(unsigned int i=0;i<times_of_next_events.size();i++)
{
	if(times_of_next_events[i]==min)
		event_id=i; 	
}
}
void execute_arrival(TRandomMersenne*rand_gen)
{
	queueing_jobs.resize(queueing_jobs.size()+1);
	
	(queueing_jobs.end()-1)->t=event_id; 

	(queueing_jobs.end()-1)->data[0]=t_now; 

	(queueing_jobs.end()-1)->data[1]=(-1.0)*log(rand_gen->Random())*means_of_processing_times[event_id]; 

	(queueing_jobs.end() - 1)->data[2] = t_now + tightness_factors[event_id] *(queueing_jobs.end()-1)->data[1];
}
void execute_production()
{
	total_production++;
	compute_time_stats();     // update statistics first                            

	machines[event_id-num_of_job_types].current_job.t=NO_JOB;         // NO_JOB means idle machine

	machines[event_id-num_of_job_types].machine_state=2;

	machines[event_id-num_of_job_types].machine_state_transition=12;
}

void event_execution(TRandomMersenne* generator) 
{       
	if (event_id >= num_of_job_types)
		execute_production();
	else
		execute_arrival(generator); 
}

void compute_dispatch_indeces(const unsigned int controller_code, unsigned int j)
{// find dispatching rule indices of all queueing jobs for machine j
			if(controller_code==0)
			{//First Come First Served rule
				for(unsigned int i=0;i<queueing_jobs.size();i++)      
					queueing_jobs[i].data[4]=queueing_jobs[i].data[0];
			}

			if(controller_code==1)
			{//Least Slack rule
				for(unsigned int i=0;i<queueing_jobs.size();i++)        
					queueing_jobs[i].data[4]=queueing_jobs[i].data[2]-t_now-queueing_jobs[i].data[1];
			}

			if(controller_code==2)
			{// Shortest Processing + Setup Time rule
				for(unsigned int i=0;i<queueing_jobs.size();i++)              
				{
					queueing_jobs[i].data[4]=queueing_jobs[i].data[1];	
					if(queueing_jobs[i].t!=machines[j].previous_job_type)
						queueing_jobs[i].data[4]+=setup_fractions[queueing_jobs[i].t]*queueing_jobs[i].data[1];
				}
			}

			if(controller_code==3)							
			{// Earliest Due Date and No Setup rule
				for(unsigned int i=0;i<queueing_jobs.size();i++)
				{
					queueing_jobs[i].data[4]=queueing_jobs[i].data[2];	
					if(queueing_jobs[i].t!=machines[j].previous_job_type)
						queueing_jobs[i].data[4]+=NEVER;
				}
			}
}

void load_job(unsigned int j)
{// select next job for processing in machine j according to dispatching rule indices

		unsigned int pos=0;                                  //the position of the next selected job
		double min=queueing_jobs[0].data[4];                 //the smallest dispathing index

		if(queueing_jobs.size()==1)
			machines[j].current_job=queueing_jobs[0];
		else
		{
			for(unsigned int i=0;i<queueing_jobs.size();i++)
			{
				if(queueing_jobs[i].data[4]<min)
				{
					min=queueing_jobs[i].data[4];
					pos=i;
				}
			}
			machines[j].current_job=queueing_jobs[pos];
		}

		if (machines[j].current_job.t!=machines[j].previous_job_type)
			machines[j].current_job.data[3]=setup_fractions[machines[j].current_job.t]*machines[j].current_job.data[1];

		queueing_jobs.erase(queueing_jobs.begin()+pos);   
			
		machines[j].machine_state=1;

		machines[j].machine_state_transition=21;      

		machines[j].previous_job_type=machines[j].current_job.t;
}

void controller(unsigned int DR_id)
{
	bool setup_machine_found=0;      // true if there is an idle machine with same setup as the selected job

	if(queueing_jobs.size()==1)
	{
		for(unsigned int i=0;i<machines.size();i++)
		{
			if( (machines[i].previous_job_type==queueing_jobs[0].t) && (machines[i].machine_state==2))
			{// if there is a single queueing job and and an idle machine with similar setup, load the job on that machine
				load_job(i);
				setup_machine_found=1;
				break;
			}
		}

		if(!setup_machine_found)
		{
			for(unsigned int i=0;i<machines.size();i++)
			{
				if( machines[i].machine_state==2)
				{// if there is a single queueing job but no idle machine with similar setup, load the job on some idle machine
						load_job(i);
						break;
				}
			}
		}
	}
	else if(queueing_jobs.size()>1)
	{
		for(unsigned int i=0;i<machines.size();i++)
		{
			if( machines[i].machine_state==2)
			{//if there are more than one queueing jobs, select one according to dispatching rule and load in on some idle machine
				compute_dispatch_indeces(DR_id, i);
				load_job(i);
				break;

			}
		}
	}
}

void update_times_of_next_events(TRandomMersenne* rand_gen) 
{
	for(unsigned int i=0;i<machines.size();i++)
	{
		if(machines[i].machine_state_transition==21)
		{// if machine i switched from idle to working, update time of next production 
			times_of_next_events[num_of_job_types+i] = t_now+machines[i].current_job.data[1] + machines[i].current_job.data[3];

			machines[i].machine_state_transition=0;
		}
		
		if(machines[i].machine_state_transition==12)
		{// if machine switched from working to idle, the event of next production is not active
			times_of_next_events[num_of_job_types+i]=NEVER;

			machines[i].machine_state_transition=0;
		}
	}

	if(event_id<num_of_job_types)
	{// if there was an arrival, update the next arrival time for that job type
		times_of_next_events[event_id] = t_now + ((-1.0)*log(rand_gen->Random())*means_of_times_between_arrivals[event_id]);
	}
}

void compute_time_stats()
{
	// update average cycle time
	double delta = t_now - machines[event_id - num_of_job_types].current_job.data[0] - mean_cycle_time;
	mean_cycle_time = (mean_cycle_time * (total_production - 1.0) + (t_now - machines[event_id - num_of_job_types].current_job.data[0])) / total_production;

	// update average earliness 
	delta = max(0.0, (machines[event_id - num_of_job_types].current_job.data[2] - t_now)) - mean_earliness;
	mean_earliness = (mean_earliness * (total_production - 1.0) + max(0.0, (machines[event_id - num_of_job_types].current_job.data[2] - t_now))) / total_production;

	// update average tardiness
	delta = max(0.0, (t_now - machines[event_id - num_of_job_types].current_job.data[2])) - mean_tardiness;
	mean_tardiness = (mean_tardiness * (total_production - 1.0) + max(0.0, (t_now - machines[event_id - num_of_job_types].current_job.data[2]))) / total_production;
}
};