﻿#include "class definitions parallel machines.cpp"

vector<double> sim_loop(time_t s1, time_t s2, unsigned int m, unsigned int n, unsigned int DR_code, const unsigned int max_production)
{//initialize random number generators
	TRandomMersenne arrival_process(s1);
	TRandomMersenne production_process(s2);

	//initialize model parameters
	parallel_machines model(m, n);                           // m is number of job types and n is number of machines 
	model.initialize_instance_from_file();                   // input file must have 4 rows and m columns
	model.initialize_times_of_next_events(&arrival_process);
	
	//run the simulation
	while (model.total_production < max_production)
	{
		model.find_time_of_next_event();
		model.event_execution(&production_process);
		model.controller(DR_code);			
		model.update_times_of_next_events(&arrival_process);
	}

	//return the statistics of simulation
	vector<double> output(3);
	output[0] = model.mean_cycle_time;
	output[1] = model.mean_earliness;	
	output[2] = model.mean_tardiness;

	return output;
}


int main()
{
	time_t seed1 = 123;
	time_t seed2 = 456;
	
	vector<double> ans = sim_loop(seed1, seed2, 3, 2, 2, 100000);

	cout << "Simulation terminated....." << endl;
	cout << "Statistics:" << endl<<endl;
	cout << "	Mean cycle time: " << ans[0] << endl;
	cout << "	Mean earliness: " << ans[1] << endl;
	cout << "	Mean tardiness: " << ans[2] << endl<<endl;
	cout << "Exiting program....." << endl << endl;
	
	return 0;
}