﻿#include "class definitions GK.cpp"

void sim_loop(time_t s[])
{
	// initialize random number generators for arrival process and production processes of a 3-stage system
	TRandomMersenne arrival_process(s[0]);
	TRandomMersenne production_process1(s[1]);
	TRandomMersenne production_process2(s[2]);
	TRandomMersenne production_process3(s[3]);

	TRandomMersenne* mersenne_array[3];
	mersenne_array[0] = &production_process1; 
	mersenne_array[1] = &production_process2; 
	mersenne_array[2] = &production_process3;

	// initialize model
	GK model(&arrival_process);

	// run the simulation
	double t_max = 1000000.0;	// maximum simulation time
	while (model.t_now < t_max)
	{
		model.find_time_of_next_event();
		model.event_execution();
		model.controller();			
		model.update_times_of_next_events(&arrival_process, mersenne_array);
	}

	cout << "Simulation terminated at time: " << model.t_now << endl;
	cout << "Statistics: " << endl;
	cout << "	Mean WIP: " << model.queue_stats.meanWIP << endl;
	cout << "	Mean finished goods inventory: " << model.queue_stats.meanI << endl;
	cout << "	Mean backorders: " << model.queue_stats.meanB << endl;
}

int main()
{
	time_t sporos[4] = { 1111, 2222, 3333, 4444 };	// seeds for RNGs
	sim_loop(sporos);
	return 0;
}